## -*-Tcl-*-
 # ###################################################################
 #  Vince's Additions - an extension package for Alpha
 # 
 #  FILE: "templates.tcl"
 #                                    created: 27/7/97 {9:29:20 pm} 
 #                                last update: 21/11/98 {2:17:59 pm} 
 #  Author: Vince Darley
 #  E-mail: <darley@fas.harvard.edu>
 #    mail: Division of Engineering and Applied Sciences, Harvard University
 #          Oxford Street, Cambridge MA 02138, USA
 #     www: <http://www.fas.harvard.edu/~darley/>
 #  
 # Copyright (c) 1997-1998  Vince Darley, all rights reserved
 # 
 # Simple template insertion mechanism.  Can be easily overridden
 # by a more sophisticated scheme providing more features, whilst
 # still working simply if desired.  Call any of these 
 # procs from the outside:
 # 
 #  elec::Insertion args  -- insert the given args
 #  elec::CenterInsertion args -- insert, then center redraw
 #  elec::Wrap left right -- wrap left and right about the selection
 #  
 #  ring::+ -- move to the next template stop
 #  ring::- -- move to the previous template stop
 #  
 # Any piece of text given to the three 'elec::' procs has a template
 # conversion done.  Text of the form 'blah' is converted to a single
 # bullet '', with 'blah' attached to it internally.  
 # 
 # A more sophisticated template package, available separately, can
 # prompt the user with 'blah' in useful ways, and creates a proper
 # template ring.
 # 
 #  modified by  rev reason
 #  -------- --- --- -----------
 #  27/7/97  VMD 1.0 original
 # ###################################################################
 ##

alpha::feature electricTab 0.1.1 global {
    set electricTab 0
} {set electricTab 1} {set electricTab 0} help {
    Enabling the 'Electric Tab' feature allows Alpha you to use
    the procedures 'Indent or Next Stop' and 'Tab or Complete' as
    any of your 'Special Keys' bindings.
    
    What these procedures do is change the behaviour of the Tab
    key to depend upon the context.  In other words, hitting Tab
    will not usually insert a Tab, rather it may indent the
    current line, or move to the next Stop mark or complete the
    current text,...
    
    Note that 'cmd-Tab' is one of the possible key-bindings
    used to complete whatever you type.  If you use cmd-Tab as
    a 'program switcher' in MacOS or other software, then you
    obviously cannot use that keybinding in Alpha, since it
    will be intercepted by the operating system.
}

namespace eval elec {}
namespace eval bind {}
namespace eval ring {}

proc alpha::useElectricTemplates {} {}

newPref variable elecStopMarker ""

## 
 # -------------------------------------------------------------------------
 #	 
 #	"elec::_Insertion" --
 #	
 #	 Insert	a piece	of text, padding on	the	left appropriately.	 The text 
 #	 should	already	be correctly indented within itself.  
 #	 
 #	 Any piece of the text of the form 'blah' is converted into a single
 #	 bullet.  A more advanced version of this procedure, available
 #	 separately, allows the use of 'blah' to prompt the user either
 #	 in the window, or status line, and makes the template stops permanent
 #	 entities so you can cycle back and forth through a template 'ring'.
 # -------------------------------------------------------------------------
 ##
proc elec::_Insertion { center args } {
    set text [join $args ""]
    set pos [getPos]
    regsub -all "\t" $text [text::Tab] text
    if {[regexp "\[\n\r\]" $text]} {
	regsub -all "\[\n\r\]" $text "\r[text::indentTo $pos]" text
    }
    if {[regexp "" $text]} {
	regsub -all "" $text [text::halfTab] text
    }
    setMark
    global elecStopMarker
    if {[regsub -all {[^]*} $text $elecStopMarker text]} {
	insertText $text
	goto $pos
	if {$center} { centerRedraw }
	# need to go to the first tab stop
	ring::+
    } else {
	insertText $text
	if {$center} { centerRedraw }
	return
    }
}


#  possible tab key bindings  #
# note: Also provided by the base Alpha system, these overide when 
# Univs Completions package is in use (these may be more intricate).

## 
 # -------------------------------------------------------------------------
 #	 
 #	"bind::IndentOrNextstop" --
 #	
 #	 Either	insert a real tab if your mode hasn't defined its electricTab
 #	 variable, or jump to the next template	stop (if we're mid-template),
 #	 or	indent the current line	correctly.
 #	 
 #	 If this proc doesn't seem to do the right thing.  Make sure you've
 #	 got 'electricTab' set correctly in your preferences!
 # -------------------------------------------------------------------------
 ##
proc bind::IndentOrNextstop {{hard 0}} {
    global electricTab
    if {$hard || !$electricTab} {
	insertActualTab 
    } else {
	if {![ring::+]} {bind::IndentLine}
    }
}

## 
 # -------------------------------------------------------------------------
 #	 
 #	"bind::TabOrComplete" --
 #	
 #	 Either	insert a real tab if your mode hasn't defined its electricTab
 #	 variable, or invoke the completion mechanism, or indent the current 
 #	 line correctly.
 #	 
 #	 If this proc doesn't seem to do the right thing.  Make sure you've
 #	 got 'electricTab' set correctly in your preferences!
 # -------------------------------------------------------------------------
 ##
proc bind::TabOrComplete {{hard 0}} {
    global electricTab
    if {$hard || !$electricTab} {
	insertActualTab 
    } else {
	bind::Completion
    }
}

proc ring::+ {} {
    global elecStopMarker
    set pos [getPos]
    if {[pos::compare $pos == [maxPos]]} { return 0 }
    set searchResult [lindex [search -s -n -f 1 -m 0 -i 1 -r 0 $elecStopMarker $pos] 0]
    if {[string length $searchResult]} {
	goto $searchResult
	deleteChar
	return 1
    } else {
	return 0
    }
}

proc ring::- {} {
    global elecStopMarker
    set pos [getPos]
    if {[pos::compare $pos == [minPos]]} { return 0 }
    set searchResult [lindex [search -s -n -f 0 -m 0 -i 1 -r 0 $elecStopMarker $pos] 0]
    if {[string length $searchResult]} {
	goto $searchResult
	deleteChar
	return 1
    } else {
	return 0
    }
}

# Removes all tab stops from the current selection (if there is one) 
# or the current document, maintaining the cursor position in the 
# latter case.
proc ring::clear {} {
    watchCursor
    set pos [getPos]
    global elecStopMarker
    if {[pos::compare [set start $pos] == [set end [selEnd]]]} {
	set start [minPos]
	set end [maxPos]
	set text1 [getText $start $pos]
	set subs1 [regsub -all $elecStopMarker $text1 {} text1]
	set text2 [getText $pos $end]
	set subs2 [regsub -all $elecStopMarker $text2 {} text2]
	append text $text1 $text2
	replaceText $start $end $text
	goto [pos::math $pos - $subs1]
	message "[expr {$subs1 + $subs2}] stops cleared"
    } else {
	set text [getText $start $end]
	set subs3 [regsub -all $elecStopMarker $text {} text]
	replaceText $start $end $text
	set end [getPos]
	select $start $end
	message "$subs3 stops cleared"
    }
}
# indicates we're a very basic ring
proc ring::type {} { return 0 }
## 
 # -------------------------------------------------------------------------
 # 
 # "elec::CenterInsertion" --
 # 
 #  Insert and then do a refresh.  Useful for large electric insertions.
 # -------------------------------------------------------------------------
 ##
proc elec::CenterInsertion {args} {
    eval elec::_Insertion 1 $args
}

## 
 # -------------------------------------------------------------------------
 # 
 # "elec::Insertion" --
 # 
 #  Just insert the electric item
 # -------------------------------------------------------------------------
 ##
proc elec::Insertion { args } {
    eval elec::_Insertion 0 $args
}
proc elec::ReplaceText { start end args } {
    deleteText $start $end 
    eval elec::_Insertion 0 $args
}

## 
 # -------------------------------------------------------------------------
 #	 
 #	"elec::Wrap" --
 #	
 #	 Currently doesn't deal	with indentation --	the	wrapper	is supposed
 #	 to	handle that.
 #	 
 #	 Returns 0/1 to indicate if there was a selection which this proc
 #	 wrapped.
 # -------------------------------------------------------------------------
 ##
proc elec::Wrap { left right {complete 0}} {
    set pos [getPos]
    set s [getSelect]
    deleteText $pos [selEnd]
    if {$s == ""} {
	elec::Insertion $left "" $right
	if {$complete} {
	    bind::Completion
	}
	return 0
    } else {
	elec::Insertion $left $s $right
	return 1
    }
}



